/**
 * @fileOverview  The model class Author with property definitions, (class-level)
 *                check methods, setter methods, and the special methods saveAll
 *                and retrieveAll
 * @authors Gerd Wagner & Juan-Francisco Reyes
 */
import { db } from "../c/initialize.mjs";
import { NoConstraintViolation, MandatoryValueConstraintViolation, RangeConstraintViolation,
  UniquenessConstraintViolation, ReferentialIntegrityConstraintViolation }
  from "../../lib/errorTypes.mjs";
/**
 * The class Author
 * @class
 * @param {object} slots - Object creation slots.
 */
class Author {
  // using a single record parameter with ES6 function parameter destructuring
  constructor ({authorId, name}) {
    // assign properties by invoking implicit setters
    this.authorId = authorId;
    this.name = name;
  };
  get authorId() {
    return this._authorId;
  };
  static checkAuthorId( id) {
    if (!id) {
      return new NoConstraintViolation();  // may be optional as an IdRef
    } else {
      // convert to integer
      id = parseInt( id);
      if (isNaN( id) || !Number.isInteger( id) || id < 1) {
        return new RangeConstraintViolation(
          "The author ID must be a positive integer!");
      } else {
        return new NoConstraintViolation();
      }
    }
  };
  static async checkAuthorIdAsId( id) {
    var constraintViolation = Author.checkAuthorId( id);
    if ((constraintViolation instanceof NoConstraintViolation)) {
      // convert to integer
      id = parseInt(id);
      if (isNaN(id)) {
        return new MandatoryValueConstraintViolation(
          "A positive integer value for the author ID is required!");
      } else {
        let authorDocSn = await db.collection("authors").doc( id.toString()).get();
        if (authorDocSn.exists) {
          constraintViolation = new UniquenessConstraintViolation(
            "There is an author record with this author ID!");
        } else {
          constraintViolation = new NoConstraintViolation();
        }
      }
    }
    return constraintViolation;
  };
  static async checkAuthorIdAsIdRef( id) {
    var constraintViolation = Author.checkAuthorId( id);
    if ((constraintViolation instanceof NoConstraintViolation) && id) {
      let authorDocSn = await db.collection("authors").doc( id.toString()).get();
      if (!authorDocSn.exists) {
        constraintViolation = new ReferentialIntegrityConstraintViolation(
          `There is no author record with ID ${id}!`);
      }
    }
    return constraintViolation;
  };
  set authorId( id) {
    var constraintViolation = Author.checkAuthorId( id);
    if (constraintViolation instanceof NoConstraintViolation) {
      this._authorId = id;
    } else {
      throw constraintViolation;
    }
  };
  get name() {
    return this._name;
  };
  set name( n) {
    /*SIMPLIFIED CODE: no validation with Author.checkName */
    this._name = n;
  };
}
/*********************************************************
 ***  Class-level ("static") storage management methods ***
 **********************************************************/
/**
 *  Conversion between a Author object and a corresponding Firestore document
 */
Author.converter = {
  toFirestore: function ( author) {
    return {
      authorId: parseInt( author.authorId),
      name: author.name
    };
  },
  fromFirestore: function (snapshot, options) {
    const data = snapshot.data( options);
    return new Author( data);
  },
};
/**
 *  Create a new author record
 */
Author.add = async function ( slots) {
  var author = null, validationResult = null;
  try {
    author = new Author( slots);
    validationResult = Author.checkAuthorIdAsId( author.authorId.toString());
    if (!validationResult instanceof NoConstraintViolation) throw validationResult;
    validationResult = Author.checkAuthorIdAsIdRef( author.authorId.toString());
    if (!validationResult instanceof NoConstraintViolation) throw validationResult;
    const authorDocRef = db.collection("authors").doc( author.authorId.toString());
    await authorDocRef.withConverter( Author.converter).set( author);
    console.log(`Author record "${author.name}" created!`);
  } catch (e) {
    console.error(`Error creating author record: ${e}`);
  }
};
/**
 * Update an author record
 */
Author.update = async function (slots) {
  const updatedSlots = {};
  let authorRec = null,
    authorDocRef = null;
  try {
    // retrieve up-to-date author record
    authorDocRef = db.collection("authors").doc(slots.authorId);
    let authorDocSn = await authorDocRef.withConverter(Author.converter).get();
    authorRec = authorDocSn.data();
  } catch (e) {
    console.error(`${e.constructor.name}: ${e.message}`);
  }
  try {
    if (authorRec.name !== slots.name) {
      /*SIMPLIFIED CODE: no validation with Author.checkName */
      updatedSlots.name = slots.name;
    }
  } catch (e) {
    console.error(`${e.constructor.name}: ${e.message}`);
  }
  let updatedProperties = Object.keys( updatedSlots);
  if (updatedProperties.length > 0) {
    // update author record
    await authorDocRef.withConverter( Author.converter).update( updatedSlots);
    console.log(`Property(ies) "${updatedProperties.toString()}" modified for author record "${slots.name}"!`);
  } else {
    console.log(`No property value changed for author record "${slots.name}"!`);
  }
};
/**
 * Delete an author record
 */
Author.destroy = async function (authorId) {
  try {
    const booksCollRef = db.collection("books"),
      authorsCollRef = db.collection("authors"),
      bookQrySn = booksCollRef.where("authorIdRefs", "array-contains", parseInt( authorId)),
      associatedBookDocSns = (await bookQrySn.get()).docs,
      authorDocRef = authorsCollRef.doc( authorId);
    // initiate batch write
    const batch = db.batch();
    for (const b of associatedBookDocSns) {
      const bookDocRef = booksCollRef.doc( b.id);
      // remove associated authorId from each Book record
      batch.update( bookDocRef, {
        authorIdRefs: firebase.firestore.FieldValue.arrayRemove( parseInt(authorId))
      });
    }
    // delete author record
    batch.delete( authorDocRef);
    batch.commit(); // finish batch write
    console.log(`Author record ${authorId} deleted!`);
  } catch (e) {
    console.error(`Error deleting author record: ${e}`);
  }
};
/**
 * Retrieve an author record
 */
Author.retrieve = async function (id) {
  try {
    const authorRec = (await db.collection("authors").doc( id.toString())
      .withConverter( Author.converter).get()).data();
    console.log(`Author record "${authorRec.name}" retrieved!`);
    return authorRec;
  } catch (e) {
    console.error(`Error retrieving author record: ${e}`);
  }
};
/**
 * Retrieve all author records
 */
Author.retrieveAll = async function () {
  let authorsCollRef = db.collection("authors");
  try {
    const authorRecords = (await authorsCollRef.withConverter( Author.converter)
      .get()).docs.map( d => d.data());
    console.log(`${authorRecords.length} author records retrieved!`);
    return authorRecords;
  } catch (e) {
    console.error(`Error retrieving all author records: ${e}`);
  }
};

export default Author;
