/**
 * @fileOverview  The model class Book with attribute definitions, (class-level)
 *                check methods, setter methods, and the special methods saveAll
 *                and retrieveAll
 * @authors Gerd Wagner & Juan-Francisco Reyes
 */
import { db } from "../c/initialize.mjs";
import Author from "./Author.mjs";
import Publisher from "./Publisher.mjs";
import { IntervalConstraintViolation, MandatoryValueConstraintViolation,
  NoConstraintViolation, PatternConstraintViolation, RangeConstraintViolation,
  UniquenessConstraintViolation} from "../../lib/errorTypes.mjs";
import { dateToTimestamp, nextYear, timestampToDate } from "../../lib/util.mjs";

/**
 * The class Book
 * @class
 */
class Book {
  // using a single record parameter with ES6 function parameter destructuring
  constructor ({isbn, title, publicationDate, authors, authorIdRefs,
                 publisher, publisher_id}) {
    // assign properties by invoking implicit setters
    this.isbn = isbn;
    this.title = title;
    this.publicationDate = publicationDate;
    // assign object references or ID references (to be converted in setter)
    this.authors = authors || authorIdRefs;
    if (publisher || publisher_id) {
      this.publisher = publisher || publisher_id;
    }
  };
  get isbn () {
    return this._isbn;
  };
  static checkIsbn( isbn) {
    if (!isbn) return new NoConstraintViolation();
    else if (typeof isbn !== "string" || isbn.trim() === "") {
      return new RangeConstraintViolation(
        "The ISBN must be a non-empty string!");
    } else if (!/\b\d{9}(\d|X)\b/.test( isbn)) {
      return new PatternConstraintViolation("The ISBN must be "+
        "a 10-digit string or a 9-digit string followed by 'X'!");
    } else {
      return new NoConstraintViolation();
    }
  };
  static async checkIsbnAsId ( isbn) {
    var validationResult = Book.checkIsbn( isbn);
    if ((validationResult instanceof NoConstraintViolation)) {
      if (!isbn) {
        validationResult = new MandatoryValueConstraintViolation(
          "A value for the ISBN must be provided!");
      } else {
        let bookDocSn = await db.collection("books").doc( isbn).get();
        if (bookDocSn.exists) {
          validationResult = new UniquenessConstraintViolation(
            `There is already a book record with ISBN ${isbn}`);
        } else {
          validationResult = new NoConstraintViolation();
        }
      }
    }
    return validationResult;
  };
  static async checkIsbnAsIdRef ( isbn) {
    var validationResult = Book.checkIsbn( isbn);
    if ((validationResult instanceof NoConstraintViolation)) {
      if (!isbn) {
        validationResult = new MandatoryValueConstraintViolation(
          "A value for the ISBN must be provided!");
      } else {
        let bookDocSn = await db.collection("books").doc( isbn).get();
        if (!bookDocSn.exists) {
          validationResult = new UniquenessConstraintViolation(
            `There is no book record with this ISBN ${isbn}!`);
        } else {
          validationResult = new NoConstraintViolation();
        }
      }
    }
    return validationResult;
  };
  set isbn( n) {
    var validationResult = Book.checkIsbn( n);
    if (validationResult instanceof NoConstraintViolation) {
      this._isbn = n;
    } else {
      throw validationResult;
    }
  };
  get title() {
    return this._title;
  };
  set title( t) {
    //SIMPLIFIED CODE: no validation with Book.checkTitle
    this._title = t;
  };
  get publicationDate() {
    return this._publicationDate;
  };
  static checkPublicationDate (publicationDate) {
    const YEAR_FIRST_BOOK = 1459,
      y = new Date( publicationDate).getFullYear();
    if (!publicationDate) {
      return new MandatoryValueConstraintViolation(
        "A value for the publication date must be provided!");
    } else if (!(typeof publicationDate === "string" &&
      /\d{4}-(0\d|1[0-2])-([0-2]\d|3[0-1])/.test( publicationDate) &&
      !isNaN( Date.parse( publicationDate)))) {
      return new PatternConstraintViolation(
        'The publication date is not well formed');
    } else if (y < YEAR_FIRST_BOOK || y > nextYear()) {
      return new IntervalConstraintViolation(
        `The value of year must be between ${YEAR_FIRST_BOOK} and next year!`);
    } else {
      return new NoConstraintViolation();
    }
  };
  set publicationDate( d) {
    let validationResult = Book.checkPublicationDate( d);
    if (validationResult instanceof NoConstraintViolation) {
      this._publicationDate = d;
    } else {
      throw validationResult;
    }
  };
  get publisher() {
    return this._publisher;
  };
  static async checkPublisher( publisher_id) {
    var validationResult = null;
    if (!publisher_id) {
      validationResult = new NoConstraintViolation();  // optional
    } else {
      // invoke foreign key constraint check
      validationResult = await Publisher.checkNameAsIdRef( publisher_id);
    }
    return validationResult;
  };
  set publisher( p) {
    this._publisher = p;
  };
  get authors() {
    return this._authors;
  };
  static async checkAuthor (author_id) {
    var validationResult = null;
    if (!author_id) {
      // author(s) are optional
      validationResult = new NoConstraintViolation();
    } else {
      // invoke foreign key constraint check
      validationResult = await Author.checkAuthorIdAsIdRef( parseInt(author_id));
    }
    return validationResult;
  };
  set authors( a) {
    this._authors = a;
  };
}

/*********************************************************
 ***  Class-level ("static") storage management methods ***
 **********************************************************/
/**
 *  Conversion between a Book object and a corresponding Firestore document
 */
Book.converter = {
  toFirestore: function ( book) {
    const data = {
      isbn: book.isbn,
      title: book.title,
      publicationDate: dateToTimestamp( book.publicationDate),
      authorIdRefs: book.authors
    };
    if (book.publisher) data.publisher_id = book.publisher;
    return data;
  },
  fromFirestore: function (snapshot, options) {
    const book = snapshot.data( options);
    const data = {
      isbn: book.isbn,
      title: book.title,
      publicationDate: timestampToDate( book.publicationDate),
      authors: book.authorIdRefs
    };
    if (book.publisher_id) data.publisher = book.publisher_id;
    return new Book( data);
  },
};
/**
 *  Create a new book record
 */
Book.add = async function (slots) {
  var validationResult = null,
    book = null;
  try {
    book = new Book( slots);
    // invoke asynchronous ID/uniqueness check
    validationResult = await Book.checkIsbnAsId( book.isbn);
    if (!(validationResult instanceof NoConstraintViolation)) throw validationResult;
    const bookDocRef = db.collection("books").doc( book.isbn);
    await bookDocRef.withConverter( Book.converter).set( book);
    console.log(`Book record "${book.isbn}" created!`);
  } catch (e) {
    console.error(`Error creating book record: ${e}`);
  }
};
/**
 * Update a book record
 */
Book.update = async function ({isbn, title, publicationDate,
    authorIdRefsToAdd, authorIdRefsToRemove, publisher_id}) {
  const updatedSlots = {};
  let validationResult = null,
    bookRec = null,
    bookDocRef = null;
  try {
    // retrieve up-to-date book record
    bookDocRef = db.collection("books").doc( isbn);
    const bookDocSn = await bookDocRef.withConverter(Book.converter).get();
    bookRec = bookDocSn.data();
  } catch (e) {
    console.error(`${e.constructor.name}: ${e.message}`);
  }
  try {
    if (bookRec.title !== title) {
      updatedSlots.title = title;
    }
    if (bookRec.publicationDate !== publicationDate) {
      validationResult = Book.checkPublicationDate( publicationDate);
      if (validationResult instanceof NoConstraintViolation) {
        updatedSlots.publicationDate = dateToTimestamp( publicationDate);
      } else {
        throw validationResult;
      }
    }
    if (publisher_id && bookRec.publisher !== publisher_id) {
      validationResult = await Book.checkPublisher( publisher_id);
      if (validationResult instanceof NoConstraintViolation) {
        updatedSlots.publisher_id = publisher_id;
      } else {
        throw validationResult;
      }
    } else if (!publisher_id && bookRec.publisher !== undefined) {
      updatedSlots.publisher_id = firebase.firestore.FieldValue.delete();
    }
    let authorIdRefs = bookRec.authors;
    if (authorIdRefsToAdd) {
      authorIdRefs = authorIdRefs.concat( authorIdRefsToAdd.map( d => +d));
    }
    if (authorIdRefsToRemove) {
      authorIdRefsToRemove = authorIdRefsToRemove.map( d => +d);
      authorIdRefs = authorIdRefs.filter( d => !authorIdRefsToRemove.includes( d));
    }
    updatedSlots.authorIdRefs = authorIdRefs;
  } catch (e) {
    console.error(`${e.constructor.name}: ${e.message}`);
  }
  let updatedProperties = Object.keys( updatedSlots);
  if (updatedProperties.length > 0) {
    // update book record
    await bookDocRef.withConverter( Book.converter).update( updatedSlots);
    console.log(`Property(ies) "${updatedProperties.toString()}" modified for book record "${isbn}"`);
  } else {
    console.log(`No property value changed for book record "${isbn}"!`);
  }
};
/**
 * Delete a book record
 */
Book.destroy = async function (isbn) {
  try {
    await db.collection("books").doc(isbn).delete();
    console.log(`Book record "${isbn}" deleted!`);
  } catch (e) {
    console.error(`Error deleting book record: ${e}`);
  }
};
/**
 * Retrieve a book record, associated authors,
 * and publisher records
 */
Book.retrieve = async function (isbn) {
  try {
    const bookRec = (await db.collection("books").doc( isbn)
      .withConverter( Book.converter).get()).data();
    if (bookRec) console.log(`Book record "${bookRec.isbn}" retrieved.`);
    return bookRec;
  }
  catch (e) {
    console.error(`Error retrieving book record: ${e}`);
  }
};
/**
 * Retrieve block of book records
 */
Book.retrieveBlock = async function (params) {
  try {
    let booksCollRef = db.collection("books");
    // set limit and order in query
    booksCollRef = booksCollRef.limit( 21);
    if (params.order) booksCollRef = booksCollRef.orderBy( params.order);
    // set pagination 'startAt' cursor
    if (params.cursor) {
      if (params.order === "publicationDate") {
        booksCollRef = booksCollRef.startAt( dateToTimestamp( params.cursor));
      }
      else booksCollRef = booksCollRef.startAt( params.cursor);
    }
    const bookRecords = (await booksCollRef.withConverter( Book.converter)
      .get()).docs.map( d => d.data());
    console.log(`Block of book records retrieved! (cursor: ${bookRecords[0][params.order]})`);
    return bookRecords;
  } catch (e) {
    console.error(`Error retrieving all book records: ${e}`);
  }
};

export default Book;
