/**
 * @fileOverview  The model class Publisher with property definitions, (class-level) check methods,
 *                setter methods, and the special methods saveAll and retrieveAll
 * @authors Gerd Wagner & Juan-Francisco Reyes
 */
import { db } from "../c/initialize.mjs";
import { NoConstraintViolation, MandatoryValueConstraintViolation, RangeConstraintViolation,
  UniquenessConstraintViolation, ReferentialIntegrityConstraintViolation }
  from "../../lib/errorTypes.mjs";
/**
 * The class Publisher
 * @class
 * @param {object} slots - Object creation slots.
 */
class Publisher {
  // using a single record parameter with ES6 function parameter destructuring
  constructor ({name, address}) {
    // assign properties by invoking implicit setters
    this.name = name;
    this.address = address;
  };
  get name() {
    return this._name;
  };
  static checkName( n) {
    if (!n) {
      return new NoConstraintViolation();  // not mandatory
    } else {
      if (typeof n !== "string" || n.trim() === "") {
        return new RangeConstraintViolation(
          "The name must be a non-empty string!");
      } else {
        return new NoConstraintViolation();
      }
    }
  };
  static async checkNameAsId( n) {
    var validationResult = Publisher.checkName( n);
    if ((validationResult instanceof NoConstraintViolation)) {
      if (!n) {
        return new MandatoryValueConstraintViolation(
          "A publisher name is required!");
      } else {
        let publisherDocSn = await db.collection("publishers").doc( n).get();
        if (publisherDocSn.exists) {
          validationResult = new UniquenessConstraintViolation(
            "There is already a publisher record with this name!");
        } else {
          validationResult = new NoConstraintViolation();
        }
      }
    }
    return validationResult;
  };
  static async checkNameAsIdRef( n) {
    var validationResult = Publisher.checkName( n);
    if ((validationResult instanceof NoConstraintViolation) && n) {
      let publisherDocSn = await db.collection("publishers").doc( n).get();
      if (!publisherDocSn.exists) {
        validationResult = new ReferentialIntegrityConstraintViolation(
          "There is no publisher record with this name!");
      }
    }
    return validationResult;
  };
  set name( n) {
    var constraintViolation = Publisher.checkName( n);
    if (constraintViolation instanceof NoConstraintViolation) {
      this._name = n;
    } else {
      throw constraintViolation;
    }
  };
  get address() {
    return this._address;
  };
  //SIMPLIFIED CODE:  Publisher.checkAddress has not been defined
  set address( a) {
    //SIMPLIFIED/MISSING CODE:  invoke Publisher.checkAddress
    this._address = a;
  };
}

/*********************************************************
 ***  Class-level ("static") storage management methods ***
 **********************************************************/
/**
 *  Conversion between a Publisher object and a corresponding Firestore document
 */
Publisher.converter = {
  toFirestore: function (publisher) { // setter
    return {
      name: publisher.name,
      address: publisher.address
    };
  },
  fromFirestore: function (snapshot, options) {
    const data = snapshot.data(options);
    return new Publisher( data);
  },
};
/**
 * Create a new publisher record
 */
Publisher.add = async function ( slots) {
  var publisher = null, validationResult = null;
  try {
    publisher = new Publisher( slots);
    validationResult = Publisher.checkNameAsId( publisher.name);
    if (!validationResult instanceof NoConstraintViolation) throw validationResult;
    validationResult = Publisher.checkNameAsIdRef( publisher.name);
    if (!validationResult instanceof NoConstraintViolation) throw validationResult;
    const publisherDocRef = db.collection("publishers").doc( publisher.name);
    await publisherDocRef.withConverter( Publisher.converter).set( publisher);
    console.log(`Publisher record "${publisher.name}" created!`);
  } catch (e) {
    console.error(`Error when adding publisher record: ${e}`);
  }
};
/**
 * Update a publisher record
 */
Publisher.update = async function ( slots) {
  const updatedSlots = {};
  let publisherRec = null,
    publisherDocRef = null;
  try {
    // retrieve up-to-date publisher record
    publisherDocRef = db.collection("publishers").doc( slots.name);
    let publisherDocSn = await publisherDocRef.withConverter( Publisher.converter).get();
    publisherRec = publisherDocSn.data();
  } catch (e) {
    console.error(`${e.constructor.name}: ${e.message}`);
  }
  try {
    if (publisherRec.address !== slots.address) {
      /*SIMPLIFIED CODE: no validation with Publisher.checkAddress */
      updatedSlots.address = slots.address;
    }
  } catch (e) {
    console.error(`${e.constructor.name}: ${e.message}`);
  }
  let updatedProperties = Object.keys( updatedSlots);
  if (updatedProperties.length > 0) {
    // update publisher record
    await publisherDocRef.update( updatedSlots);
    console.log(`Property(ies) "${updatedProperties.toString()}" modified for publisher record "${slots.name}"`);
  } else {
    console.log(`No property value changed for publisher record "${slots.name}"!`);
  }
};
/**
 * Delete a publisher record
 */
Publisher.destroy = async function (name) {
  try {
    const booksCollRef = db.collection("books"),
      publishersCollRef = db.collection("publishers"),
      bookQrySn = booksCollRef.where("publisher_id", "==", name),
      associatedBookDocSns = (await bookQrySn.get()).docs,
      publisherDocRef = publishersCollRef.doc( name);
    // initiate batch write
    const batch = db.batch();
    for (const b of associatedBookDocSns) {
      const bookDocRef = booksCollRef.doc( b.id);
      // remove associated publisher from each book record
      batch.update( bookDocRef, {
        publisher_id: firebase.firestore.FieldValue.delete()
      });
    }
    // delete publisher record
    batch.delete( publisherDocRef);
    batch.commit(); // finish batch write
    console.log(`Publisher record "${name}" deleted!`);
  } catch (e) {
    console.error(`Error deleting publisher record: ${e}`);
  }
};
/**
 * Retrieve an publisher record
 */
Publisher.retrieve = async function (name) {
  try {
    const publisherRec = (await db.collection("publishers").doc( name)
      .withConverter( Publisher.converter).get()).data();
    console.log(`Publisher record "${publisherRec.name}" retrieved.`);
    return publisherRec;
  } catch (e) {
    console.error(`Error retrieving publisher record: ${e}`);
  }
};
/**
 * Retrieve all publisher records
 */
Publisher.retrieveAll = async function () {
  let publishersCollRef = db.collection("publishers");
  try {
    const publisherRecords = (await publishersCollRef.withConverter( Publisher.converter)
      .get()).docs.map( d => d.data());
    console.log(`${publisherRecords.length} publisher records retrieved!`);
    return publisherRecords;
  } catch (e) {
    console.error(`Error retrieving all publisher records: ${e}`);
  }
};

export default Publisher;
