/**
 * @fileOverview  Contains various view functions for managing authors
 * @authors Gerd Wagner & Juan-Francisco Reyes
 */
/***************************************************************
 Import classes, datatypes and utility procedures
 ***************************************************************/
import Author from "../m/Author.mjs";
import { showProgressBar } from "../../lib/util.mjs";

/***************************************************************
 Declare variables for accessing UI elements
 ***************************************************************/
// use-case-independent UI elements
const menuButtonEls = document.querySelectorAll(
  "section#Author-M > ul > li > button"),
  backButtonEls = document.querySelectorAll("button.back-to-menu"),
  tableBodyEl = document.querySelector("section#Author-R > table > tbody"),
  createFormEl = document.querySelector("section#Author-C > form"),
  updateFormEl = document.querySelector("section#Author-U > form"),
  deleteFormEl = document.querySelector("section#Author-D > form"),
  authorMSectionEl = document.getElementById("Author-M"),
  authorRSectionEl = document.getElementById("Author-R"),
  authorCSectionEl = document.getElementById("Author-C"),
  authorUSectionEl = document.getElementById("Author-U"),
  authorDSectionEl = document.getElementById("Author-D");

/***************************************************************
 Add event listeners
 ***************************************************************/
// set up back-to-menu buttons for all use cases
for (let btn of backButtonEls) {
  btn.addEventListener('click', function () {refreshManageDataUI();});
}
// neutralize the submit event for all use cases
for (let form of [createFormEl, updateFormEl, deleteFormEl]) {
  form.addEventListener("submit", function (e) {
    e.preventDefault();
  });
}

/**********************************************
 * Use case Retrieve and List All Authors
 **********************************************/
menuButtonEls[0].addEventListener("click", async function () {
  tableBodyEl.innerHTML = "";
  authorMSectionEl.hidden = true;
  authorRSectionEl.hidden = false;
  showProgressBar( "Author-R","show");
  const authorRecords = await Author.retrieveAll();
  for (let authorRec of authorRecords) {
    const row = tableBodyEl.insertRow();
    row.insertCell().textContent = authorRec.authorId;
    row.insertCell().textContent = authorRec.name;
  }
  showProgressBar( "Author-R","hide");
});

/**********************************************
 * Use case Create Author
 **********************************************/
menuButtonEls[1].addEventListener("click", function () {
  authorMSectionEl.hidden = true;
  authorCSectionEl.hidden = false;
  createFormEl.reset();
});
// set up event handlers for responsive constraint validation
createFormEl["authorId"].addEventListener("input", async function () {
  let responseValidation = await Author.checkAuthorIdAsId( createFormEl.authorId.value);
  createFormEl["authorId"].setCustomValidity( responseValidation.message);
});
/* SIMPLIFIED CODE: no responsive validation of name */

// handle Save button click events
createFormEl["commit"].addEventListener("click", async function () {
  const slots = {
    authorId: createFormEl.authorId.value,
    name: createFormEl.name.value
  };
  showProgressBar( "Author-C","show");
  // check all input fields and show error messages
  createFormEl["authorId"].setCustomValidity((
    await Author.checkAuthorIdAsId( slots.authorId)).message);
  /* SIMPLIFIED CODE: no before-submit validation of name */
  // save the input data only if all form fields are valid
  if (createFormEl.checkValidity()) {
    await Author.add( slots);
    createFormEl.reset();
  }
  showProgressBar( "Author-C","hide");
});

/**********************************************
 * Use case Update Author
 **********************************************/
menuButtonEls[2].addEventListener("click", async function () {
  authorMSectionEl.hidden = true;
  authorUSectionEl.hidden = false;
  updateFormEl.reset();
});
/**
 * handle author ID input: when an author ID is entered and the user changes
 * focus to the next form element, populate the form with the author data
 */
updateFormEl["authorId"].addEventListener("input", async function () {
  let responseValidation = await Author.checkAuthorIdAsIdRef( updateFormEl["authorId"].value);
  updateFormEl["authorId"].setCustomValidity( responseValidation.message);
  updateFormEl["commit"].disabled = responseValidation.message;
  if (!updateFormEl["authorId"].value) updateFormEl.reset();
});
updateFormEl["authorId"].addEventListener("blur", async function () {
  if (updateFormEl["authorId"].checkValidity() && updateFormEl["authorId"].value) {
    const authorRec = await Author.retrieve( updateFormEl["authorId"].value);
    updateFormEl["name"].value = authorRec.name;
  } else updateFormEl.reset();
});

// handle Save button click events
updateFormEl["commit"].addEventListener("click", async function () {
  if (!updateFormEl["authorId"].value) return;
  const slots = {
    authorId: updateFormEl.authorId.value,
    name: updateFormEl.name.value
  }
  // check all property constraints
  /* SIMPLIFIED CODE: no before-save validation of name */
  // save the input data only if all of the form fields are valid
  if (updateFormEl.checkValidity()) {
    await Author.update( slots);
    updateFormEl.reset();
  }
});

/**********************************************
 * Use case Delete Author
 **********************************************/
menuButtonEls[3].addEventListener("click", async function () {
  authorMSectionEl.hidden = true;
  authorDSectionEl.hidden = false;
  deleteFormEl.reset();
});
/**
 * handle author ID input: when an author ID is entered and the user changes
 * focus to the next form element, populate the form with the author data
 */
deleteFormEl["authorId"].addEventListener("input", async function () {
  let responseValidation = await Author.checkAuthorIdAsIdRef( deleteFormEl["authorId"].value);
  deleteFormEl["authorId"].setCustomValidity( responseValidation.message);
  deleteFormEl["commit"].disabled = responseValidation.message;
  if (!deleteFormEl["authorId"].value) deleteFormEl.reset();
});

// handle Delete button click events
deleteFormEl["commit"].addEventListener("click", async function () {
  const authorIdRef = deleteFormEl["authorId"].value;
  if (deleteFormEl["authorId"].checkValidity() && authorIdRef) {
    const authorRec = await Author.retrieve( authorIdRef);
    deleteFormEl["name"].value = authorRec.name;
  }
  if (confirm( "Do you really want to delete this author record?")) {
    await Author.destroy( authorIdRef);
    deleteFormEl.reset();
  }
});

/**********************************************
 * Refresh the Manage Authors Data UI
 **********************************************/
function refreshManageDataUI() {
  // show the manage author UI and hide the other UIs
  authorMSectionEl.hidden = false;
  authorRSectionEl.hidden = true;
  authorCSectionEl.hidden = true;
  authorUSectionEl.hidden = true;
  authorDSectionEl.hidden = true;
}

/** Retrieve data and setup the author management UI */
// Set up Manage Authors UI
refreshManageDataUI();
