/**
 * @fileOverview  Contains various view functions for managing books
 * @authors Gerd Wagner & Juan-Francisco Reyes
 */
/***************************************************************
 Import classes, datatypes and utility procedures
 ***************************************************************/
import Book from "../m/Book.mjs";
import { createMultiSelectionWidget, showProgressBar } from "../../lib/util.mjs";

/***************************************************************
 Declare variables for accessing UI elements
 ***************************************************************/
// use-case-independent UI elements
const menuButtonEls = document.querySelectorAll(
  "section#Book-M > ul > li > button"),
  backButtonEls = document.querySelectorAll("button.back-to-menu"),
  formEls = document.querySelectorAll("section > form"),
  bookMSectionEl = document.getElementById("Book-M"),
  bookRSectionEl = document.getElementById("Book-R"),
  bookCSectionEl = document.getElementById("Book-C"),
  bookUSectionEl = document.getElementById("Book-U"),
  bookDSectionEl = document.getElementById("Book-D"),
  tableBodyEl = bookRSectionEl.querySelector("table > tbody"),
  createFormEl = bookCSectionEl.querySelector("form"),
  updateFormEl = bookUSectionEl.querySelector("form"),
  deleteFormEl = bookDSectionEl.querySelector("form");

/***************************************************************
 Set up general, use-case-independent event listeners
 ***************************************************************/
// set up back-to-menu buttons for all use cases
for (let btn of backButtonEls) {
  btn.addEventListener('click', function () {refreshManageDataUI();});
}
// neutralize the submit event for all use cases
for (let frm of formEls) {
  frm.addEventListener("submit", function (e) {
    e.preventDefault();
  });
}

/**********************************************
 * Use case Retrieve/List All Books
 **********************************************/
// initialize pagination mapping references
let cursor = null,
  previousPageRef = null,
  nextPageRef = null,
  startAtRefs = [];
let order = "isbn"; // default order value
const selectOrderEl = bookRSectionEl.querySelector("div > label > select"),
  previousBtnEl = document.getElementById("previousPage"),
  nextBtnEl = document.getElementById("nextPage");
/**
 * event handler for main menu button
 */
menuButtonEls[0].addEventListener("click", async function () {
  bookMSectionEl.hidden = true;
  bookRSectionEl.hidden = false;
  await createBlock();
  startAtRefs.push( cursor); // set 'first' startAt page reference
  previousBtnEl.disabled = true;
});
/**
 * create listing page
 */
async function createBlock (startAt) {
  tableBodyEl.innerHTML = "";
  showProgressBar( "Book-R","show");
  const bookRecords = await Book.retrieveBlock({"order": order, "cursor": startAt});
  // set page references for current (cursor) page
  cursor = bookRecords[0][order];
  // set next startAt page reference, if not next page, assign 'null' value
  nextPageRef = (bookRecords.length < 21) ? null : bookRecords[bookRecords.length -1][order];
  for (let bookRec of bookRecords) {
    const row = tableBodyEl.insertRow(-1);
    row.insertCell(-1).textContent = bookRec.isbn;
    row.insertCell(-1).textContent = bookRec.title;
    row.insertCell(-1).textContent = bookRec.publicationDate;
    row.insertCell(-1).textContent = bookRec.authors;
    row.insertCell(-1).textContent = bookRec.publisher;
  }
  showProgressBar("Book-R", "hide");
}
/**
 * 'Previous' button
 */
previousBtnEl.addEventListener("click", async function () {
  // locate current page reference in index of page references
  previousPageRef = startAtRefs[startAtRefs.indexOf( cursor) - 1];
  // create new page
  await createBlock( previousPageRef);
  // disable 'previous' button if cursor is first page
  if (cursor === startAtRefs[0]) previousBtnEl.disabled = true;
  // enable 'next' button if cursor is not last page
  if (cursor !== startAtRefs[startAtRefs.length -1]) nextBtnEl.disabled = false;
});
/**
 *  'Next' button
 */
nextBtnEl.addEventListener("click", async function () {
  await createBlock( nextPageRef);
  // add new page reference if not present in index
  if (!startAtRefs.find( i => i === cursor)) startAtRefs.push( cursor);
  // disable 'next' button if cursor is last page
  if (!nextPageRef) nextBtnEl.disabled = true;
  // enable 'previous' button if cursor is not first page
  if (cursor !== startAtRefs[0]) previousBtnEl.disabled = false;
});
/**
 * handle order selection events: when an order is selected,
 * populate the list according to the selected order
 */
selectOrderEl.addEventListener("change", async function (e) {
  order = e.target.value;
  startAtRefs = [];
  await createBlock();
  startAtRefs.push( cursor);
  previousBtnEl.disabled = true;
  nextBtnEl.disabled = false;
});

/**********************************************
 * Use case Create Book
 **********************************************/
const authorsCrtWidget = createFormEl.querySelector(
  "section#Book-C > form > div > .MultiSelectionWidget");
menuButtonEls[1].addEventListener("click", async function () {
  bookMSectionEl.hidden = true;
  bookCSectionEl.hidden = false;
  createMultiSelectionWidget( authorsCrtWidget, [],
    "crtAuthors", "Enter ID", 1);
  createFormEl.reset();
});
// set up event handlers for responsive constraint validation
createFormEl.isbn.addEventListener("input", async function () {
  let responseValidation = await Book.checkIsbnAsId( createFormEl.isbn.value);
  createFormEl["isbn"].setCustomValidity( responseValidation.message);
});
/* SIMPLIFIED/MISSING CODE: add event listeners for responsive
validation on user input with Book.checkTitle */

// handle Save button click events
createFormEl["commit"].addEventListener("click", async function () {
  const authorsListEl = authorsCrtWidget.querySelector("ul");
  const slots = {
    isbn: createFormEl["isbn"].value,
    title: createFormEl["title"].value,
    publicationDate: createFormEl["publicationDate"].value,
    publisher_id: createFormEl["publisher"].value,
    authorIdRefs: [],
  };
  // check all input fields and show error messages
  createFormEl["isbn"].setCustomValidity(( await Book.checkIsbnAsId( slots.isbn)).message);
  /* SIMPLIFIED CODE: no before-submit validation of title */
  createFormEl["publicationDate"].setCustomValidity(
    Book.checkPublicationDate( slots.publicationDate).message);
  let responseValidation = await Book.checkPublisher( slots.publisher_id);
  createFormEl["publisher"].setCustomValidity( responseValidation.message);
  // get the list of selected authors
  for (const el of authorsListEl.children) {
    slots.authorIdRefs.push( parseInt( el.getAttribute("data-value")));
  }
  if (slots.authorIdRefs.length > 0) {
    for (const a of slots.authorIdRefs) {
      let responseValidation = await Book.checkAuthor( a);
      if (responseValidation.message) {
        createFormEl["crtAuthors"].setCustomValidity( responseValidation.message);
        break;
      }  else createFormEl["crtAuthors"].setCustomValidity( "");
    }
  } else {
    createFormEl["crtAuthors"].setCustomValidity(
      createFormEl["crtAuthors"].value ? "" : "No author selected!"
    );
  }
  showProgressBar("Book-C", "show");
  // save the input data only if all form fields are valid
  if (createFormEl.checkValidity()) {
    await Book.add( slots);
    // drop widget content
    authorsListEl.innerHTML = "";
    createFormEl.reset();
  }
  showProgressBar( "Book-C", "hide");
});

/**********************************************
 * Use case Update Book
 **********************************************/
const authorsUptWidget = updateFormEl.querySelector(
  "section#Book-U > form > div > .MultiSelectionWidget");
menuButtonEls[2].addEventListener("click", async function () {
  bookMSectionEl.hidden = true;
  bookUSectionEl.hidden = false;
  updateFormEl.reset();
});
/**
 * handle book ISBN input: when a book ISBN is entered and the user changes
 * focus to the next form element, populate the form with the book data
 */
let bookRec = null;
updateFormEl["isbn"].addEventListener("input", async function () {
  let responseValidation = await Book.checkIsbnAsIdRef( updateFormEl["isbn"].value);
  if (responseValidation) updateFormEl["isbn"].setCustomValidity( responseValidation.message);
  if (!updateFormEl["isbn"].value) {
    updateFormEl.reset();
    // drop widget content
    const authorsListEl = authorsUptWidget.querySelector("ul");
    authorsListEl.innerHTML = "";
  }
});
updateFormEl["isbn"].addEventListener("blur", async function () {
  if (updateFormEl["isbn"].checkValidity() && updateFormEl["isbn"].value) {
    bookRec = await Book.retrieve( updateFormEl["isbn"].value);
    updateFormEl["isbn"].value = bookRec.isbn;
    updateFormEl["title"].value = bookRec.title;
    updateFormEl["publicationDate"].value = bookRec.publicationDate;
    if (bookRec.publisher) updateFormEl["publisher"].value = bookRec.publisher;
    createMultiSelectionWidget( authorsUptWidget, bookRec.authors,
      "uptAuthors", "Enter ID", 1);
  } else {
    updateFormEl.reset();
    // drop widget content
    const authorsListEl = authorsUptWidget.querySelector("ul");
    authorsListEl.innerHTML = "";
  }
});
// handle Save button click events
updateFormEl["commit"].addEventListener("click", async function () {
  const authorsListEl = authorsUptWidget.querySelector("ul");
  if (!updateFormEl["isbn"].value) return;
  const slots = {
    isbn: updateFormEl["isbn"].value,
    title: updateFormEl["title"].value,
    publicationDate: updateFormEl["publicationDate"].value,
    publisher_id: updateFormEl["publisher"].value
  };
  // construct authorIdRefs-ToAdd/ToRemove lists
  const authorIdRefsToAdd = [], authorIdRefsToRemove = [];
  for (const authorItemEl of authorsListEl.children) {
    if (authorItemEl.classList.contains("removed")) {
      authorIdRefsToRemove.push(authorItemEl.getAttribute("data-value"));
    }
    if (authorItemEl.classList.contains("added")) {
      authorIdRefsToAdd.push(authorItemEl.getAttribute("data-value"));
    }
  }
  // if the add/remove list is non-empty, create a corresponding slot
  if (authorIdRefsToRemove.length > 0) {
    slots.authorIdRefsToRemove = authorIdRefsToRemove;
  }
  if (authorIdRefsToAdd.length > 0) {
    slots.authorIdRefsToAdd = authorIdRefsToAdd;
  }
  /* MISSING CODE */
  // add event listeners for responsive validation
  if (slots.authorIdRefsToAdd) {
    for (const a of slots.authorIdRefsToAdd) {
      let responseValidation = await Book.checkAuthor( a);
      if (responseValidation.message !== "") {
        updateFormEl["uptAuthors"].setCustomValidity( responseValidation.message);
        break;
      } else updateFormEl["uptAuthors"].setCustomValidity("");
    }
  }
  /* SIMPLIFIED CODE: no before-submit validation of title */
  updateFormEl["publicationDate"].setCustomValidity(
    Book.checkPublicationDate( slots.publicationDate).message);
  let responseValidation = await Book.checkPublisher( slots.publisher_id);
  updateFormEl["publisher"].setCustomValidity( responseValidation.message);

  // commit the update only if all form field values are valid
  if (updateFormEl.checkValidity()) {
    await Book.update( slots);
    // drop widget content
    authorsListEl.innerHTML = "";
    updateFormEl.reset();
  }
});

/**********************************************
 * Use case Delete Book
 **********************************************/
menuButtonEls[3].addEventListener("click", async function () {
  bookMSectionEl.hidden = true;
  bookDSectionEl.hidden = false;
  deleteFormEl.reset();
});
/**
 * handle book ISBN input: when a book ISBN is entered and the user presses the
 * delete button, populate the form with the book data and ask for confirm
 */
deleteFormEl["isbn"].addEventListener("input", async function () {
  let responseValidation = await Book.checkIsbnAsIdRef( deleteFormEl["isbn"].value);
  deleteFormEl["isbn"].setCustomValidity( responseValidation.message);
});

// handle Delete button click events
deleteFormEl["commit"].addEventListener("click", async function () {
  const bookIdRef = deleteFormEl["isbn"].value;
  if (!bookIdRef) return;
  if (deleteFormEl["isbn"].checkValidity() && bookIdRef) {
    const bookRec = await Book.retrieve( bookIdRef);
    deleteFormEl["isbn"].value = bookRec.isbn;
    deleteFormEl["title"].value = bookRec.title;
    deleteFormEl["publicationDate"].value = bookRec.publicationDate;
    if (bookRec.publisher) deleteFormEl["publisher"].value = bookRec.publisher;
    deleteFormEl["authors"].value = bookRec.authors;
  }
  if (confirm("Do you really want to delete this book?")) {
    await Book.destroy( bookIdRef);
    deleteFormEl.reset();
  }
});

/**********************************************
 * Refresh the Manage Books Data UI
 **********************************************/
function refreshManageDataUI() {
  // show the manage book UI and hide the other UIs
  bookMSectionEl.hidden = false;
  bookRSectionEl.hidden = true;
  bookCSectionEl.hidden = true;
  bookUSectionEl.hidden = true;
  bookDSectionEl.hidden = true;
}

/** Retrieve data and setup the book management UI */
// set up Manage Book UI
refreshManageDataUI();
