/**
 * @fileOverview  Contains various view functions for managing publishers
 * @authors Gerd Wagner & Juan-Francisco Reyes
 */
/***************************************************************
 Import classes, datatypes and utility procedures
 ***************************************************************/
import Publisher from "../m/Publisher.mjs";
import { showProgressBar } from "../../lib/util.mjs";

/***************************************************************
 Declare variables for accessing UI elements
 ***************************************************************/
// use-case-independent UI elements
const menuButtonEls = document.querySelectorAll(
  "section#Publisher-M > ul > li > button"),
  backButtonEls = document.querySelectorAll("button.back-to-menu"),
  formEls = document.querySelectorAll("section > form"),
  tableBodyEl = document.querySelector("section#Publisher-R > table > tbody"),
  createFormEl = document.querySelector("section#Publisher-C > form"),
  updateFormEl = document.querySelector("section#Publisher-U > form"),
  deleteFormEl = document.querySelector("section#Publisher-D > form"),
  publisherMSectionEl = document.getElementById("Publisher-M"),
  publisherRSectionEl = document.getElementById("Publisher-R"),
  publisherCSectionEl = document.getElementById("Publisher-C"),
  publisherUSectionEl = document.getElementById("Publisher-U"),
  publisherDSectionEl = document.getElementById("Publisher-D");

/***************************************************************
 Add event listeners
 ***************************************************************/
// set up back-to-menu buttons for all use cases
for (let btn of backButtonEls) {
  btn.addEventListener('click', function () {refreshManageDataUI();});
}
// neutralize the submit event for all use cases
for (let form of formEls) {
  form.addEventListener("submit", function (e) {
    e.preventDefault();
  });
}

/**********************************************
 * Use case Retrieve/List All Publishers
 **********************************************/
menuButtonEls[0].addEventListener("click", async function () {
  tableBodyEl.innerHTML = "";
  publisherMSectionEl.hidden = true;
  publisherRSectionEl.hidden = false;
  showProgressBar( "Publisher-R","show");
  const publisherRecords = await Publisher.retrieveAll();
  for (const publisher of publisherRecords) {
    const row = tableBodyEl.insertRow();
    row.insertCell().textContent = publisher.name;
    row.insertCell().textContent = publisher.address;
  }
  showProgressBar( "Publisher-R","hide");
});

/**********************************************
 * Use case Create Publisher
 **********************************************/
menuButtonEls[1].addEventListener("click", function () {
  publisherMSectionEl.hidden = true;
  publisherCSectionEl.hidden = false;
  createFormEl.reset();
});
createFormEl.name.addEventListener("input", async function () {
  let responseValidation = await Publisher.checkNameAsId( createFormEl.name.value);
  createFormEl.name.setCustomValidity( responseValidation.message);
});
/* SIMPLIFIED CODE: no responsive validation of address */

// handle Save button click events
createFormEl["commit"].addEventListener("click", async function () {
  const slots = {
    name: createFormEl.name.value,
    address: createFormEl.address.value
  };
  showProgressBar( "Publisher-C","show");
  // check all input fields and show error messages
  createFormEl.name.setCustomValidity(( await Publisher.checkNameAsId( slots.name)).message);
  /* SIMPLIFIED CODE: no before-submit validation of name */
  // save the input data only if all form fields are valid
  if (createFormEl.checkValidity()) {
    await Publisher.add( slots);
    createFormEl.reset();
  }
  showProgressBar( "Publisher-C","hide");
});

/**********************************************
 * Use case Update Publisher
 **********************************************/
menuButtonEls[2].addEventListener("click", async function () {
  publisherMSectionEl.hidden = true;
  publisherUSectionEl.hidden = false;
  updateFormEl.reset();
});
/**
 * handle publisher name input: when a publisher name is entered and the user changes
 * focus to the next form element, populate the form with the publisher data
 */
updateFormEl["name"].addEventListener("input", async function () {
  let responseValidation = await Publisher.checkNameAsIdRef( updateFormEl["name"].value);
  updateFormEl["name"].setCustomValidity( responseValidation.message);
  updateFormEl["commit"].disabled = responseValidation.message;
  if (!updateFormEl["name"].value) updateFormEl.reset();
});
updateFormEl["name"].addEventListener("blur", async function () {
  if (updateFormEl["name"].checkValidity() && updateFormEl["name"].value) {
    const publisherRec = await Publisher.retrieve( updateFormEl["name"].value);
    updateFormEl["name"].value = publisherRec.name;
    updateFormEl["address"].value = publisherRec.address;
  } else updateFormEl.reset();
});

// handle Save button click events
updateFormEl["commit"].addEventListener("click", async function () {
  if (!updateFormEl["name"].value) return;
  const slots = {
    name: updateFormEl.name.value,
    address: updateFormEl.address.value
  }
  // check all property constraints
  /* SIMPLIFIED CODE: no before-save validation of name */
  // save the input data only if all of the form fields are valid
  if (updateFormEl.checkValidity()) {
    await Publisher.update( slots);
    updateFormEl.reset();
  }
});

/**********************************************
 * Use case Delete Publisher
 **********************************************/
menuButtonEls[3].addEventListener("click", async function () {
  publisherMSectionEl.hidden = true;
  publisherDSectionEl.hidden = false;
  deleteFormEl.reset();
});
/**
 * handle publisher name input: when a publisher name is entered and the user changes
 * focus to the next form element, populate the form with the publisher data
 */
deleteFormEl["name"].addEventListener("input", async function () {
  let responseValidation = await Publisher.checkNameAsIdRef( deleteFormEl["name"].value);
  deleteFormEl["name"].setCustomValidity( responseValidation.message);
  deleteFormEl["commit"].disabled = responseValidation.message;
  if (!deleteFormEl["name"].value) deleteFormEl.reset();
});

// handle Delete button click events
deleteFormEl["commit"].addEventListener("click", async function () {
  const publisherNameRef = deleteFormEl["name"].value;
  if (deleteFormEl["name"].checkValidity() && publisherNameRef) {
    const publisherRec = await Publisher.retrieve( publisherNameRef);
    deleteFormEl["name"].value = publisherRec.name;
    deleteFormEl["address"].value = publisherRec.address;
  }
  if (confirm( "Do you really want to delete this publisher?")) {
    await Publisher.destroy( publisherNameRef);
    deleteFormEl.reset();
  }
});

/**********************************************
 * Refresh the Manage Publishers Data UI
 **********************************************/
function refreshManageDataUI() {
  // show the manage publisher UI and hide the other UIs
  publisherMSectionEl.hidden = false;
  publisherRSectionEl.hidden = true;
  publisherCSectionEl.hidden = true;
  publisherUSectionEl.hidden = true;
  publisherDSectionEl.hidden = true;
}

/** Retrieve data and setup the publisher management UI */
// Set up Manage Publishers UI
refreshManageDataUI();
