/**
 * @fileOverview  Defines utility procedures/functions
 * @author Gerd Wagner and Juan-Francisco Reyes
 */

/**
 * Verifies if a value represents an integer
 * @param {number} x
 * @return {boolean}
 */
function isNonEmptyString (x) {
  return typeof(x) === "string" && x.trim() !== "";
}
/**
 * Serialize a Date object as an ISO date string
 * @return  YYYY-MM-DD
 */
function createIsoDateString(d) {
  return d.toISOString().substring(0,10);
}
/**
 * Return the next year value (e.g. if now is 2013 the function will return 2014)
 * @return {number}  the integer representing the next year value
 */
function nextYear() {
  const date = new Date();
  return (date.getFullYear() + 1);
}

// *************** D O M - Related ****************************************
/**
 * Create a Push Button
 * @param {string} txt [optional]
 * @return {object}
 */
function createPushButton( txt) {
  var pB = document.createElement("button");
  pB.type = "button";
  if (txt) pB.textContent = txt;
  return pB;
}
/**
 * Create a DOM option element
 *
 * @param {string} val
 * @param {string} txt
 * @param {string} classValues [optional]
 *
 * @return {object}
 */
function createOption( val, txt, classValues) {
  var el = document.createElement("option");
  el.value = val;
  el.text = txt;
  if (classValues) el.className = classValues;
  return el;
}
/**
 * Create a time element from a Date object
 *
 * @param {object} d
 * @return {object}
 */
function createTimeElem(d) {
  var tEl = document.createElement("time");
  tEl.textContent = d.toLocaleDateString();
  tEl.datetime = d.toISOString();
  return tEl;
}
/**
 * Create a list element from an map of objects
 *
 * @param {object} et  An entity table
 * @param {string} displayProp  The object property to be displayed in the list
 * @return {object}
 */

/**
 * * Convert a map of values into a UL element
 * */
function createListFromMap(et, displayProp) {
  var listEl = document.createElement("ul");
  fillListFromMap( listEl, et, displayProp);
  return listEl;
}
/**
 * Fill a list element with items from an entity table
 *
 * @param {object} listEl  A list element
 * @param {object} et  An entity table
 * @param {string} displayProp  The object property to be displayed in the list
 */
function fillListFromMap(listEl, et, displayProp) {
  const keys = Object.keys( et);
  // delete old contents
  listEl.innerHTML = "";
  // create list items from object property values
  for (let key of keys) {
    var listItemEl = document.createElement("li");
    listItemEl.textContent = et[key][displayProp];
    listEl.appendChild( listItemEl);
  }
}
/**
 * Fill a select element with option elements created from a
 * map of objects
 *
 * @param {object} selectEl  A select(ion list) element
 * @param {array} selectionRange  An array list
 * @param {object} optPar [optional]  An optional parameter record including
 *     optPar.displayProp, optPar.valueProp and optPar.selection
 */
function fillSelectWithOptions( selectEl, selectionRange, optPar) {
  // create option elements from array key and values
  const options = selectionRange.entries();
  // delete old contents
  selectEl.innerHTML = "";
  // create "no selection yet" entry
  if (!selectEl.multiple && optPar.blank) selectEl.add( createOption(""," --- "));
  //
  for (const [index, o] of options) {
    const key = index + 1;
    const optionEl = createOption(
      optPar ? (o[optPar.valueProp] ? o[optPar.valueProp] : key) : key,
      optPar ? (o[optPar.displayProp] ? o[optPar.displayProp] : o) : o
    );
    if (selectEl.multiple && optPar && optPar.selection &&
      optPar.selection.includes(key)) {
      // flag the option element with this value as selected
      optionEl.selected = true;
    } else if (optPar.selection === o[optPar.valueProp]  ) {
      optionEl.selected = true;
    }
    selectEl.add( optionEl);
  }
}
// *************** Multiple Selection Widget ****************************************
/**
 * Create the contents of an Multiple Selection widget, which is a div containing
 * 1) an input field, where item to be deleted are entered,
 * 2) a div containing a select element and an add button allowing to add a selected item
 *    to the association list
 *
 * @param {object} widgetContainerEl  The widget's container div
 * @param {object} selectionRange  An map of objects, which is used to
 *                 create the options of the select element
 * @param {string} inputTextId  input element's name attribute
 * @param {string} placeholder  input element's placeholder message
 * @param {number} minCard  value of minimal number of choices
 */
function createMultiSelectionWidget(widgetContainerEl, selectionRange,
       inputTextId, placeholder, minCard) {
  const selectedItemsListEl = document.createElement("ul");
  let el = null;
  if (!minCard) minCard = 0;  // default
  widgetContainerEl.innerHTML = ""; // delete old contents
  // event handler for removing an item from the selection
  selectedItemsListEl.addEventListener( 'click', function (e) {
    if (e.target.tagName === "BUTTON") {  // delete/undo button was clicked
      const btnEl = e.target,
        listItemEl = btnEl.parentNode;
      let listEl = listItemEl.parentNode;
      if (listEl.children.length <= minCard) {
        alert(`A record must have at least one item associated!`);
        return;
      }
      if (listItemEl.classList.contains("removed")) {
        // undoing a previous removal
        listItemEl.classList.remove("removed");
        // change button text
        btnEl.textContent = "✕";
      } else if (listItemEl.classList.contains("added")) {
        // removing a previously added item means moving it back to the selection range
        listItemEl.parentNode.removeChild( listItemEl);
      } else {
        // removing an ordinary item
        listItemEl.classList.add("removed");
        // change button text
        btnEl.textContent = "undo";
      }
    }
  });
  for (const authorId of selectionRange) {
    addItemToListOfSelectedItems( selectedItemsListEl, authorId,
      `> ${authorId}`);
  }
  // embed input text field
  const spanEl = document.querySelector(`label[for="${inputTextId}"] > span`);
  spanEl.innerHTML = "";
  const inputEl = document.createElement("input");
  inputEl.setAttribute("size", "15") ;
  inputEl.setAttribute("placeholder", placeholder);
  inputEl.setAttribute("name",inputTextId)
  widgetContainerEl.appendChild( selectedItemsListEl);
  spanEl.appendChild( inputEl);
  const addButton = createPushButton("add")
  spanEl.appendChild( addButton);
  let enteredValues = selectionRange.length > 0 ? [] : selectionRange;
  // event handler for moving an item from the input text to the selected items list
  addButton.parentNode.addEventListener( 'click', function (e) {
    if (e.target.tagName === "BUTTON") {  // add button was clicked
      if (inputEl.value) {
        if (selectionRange.includes( parseInt( inputEl.value))) { // check uniqueness
          alert(`There is an item entered with the same value: ${inputEl.value}`);
        } else if (isNaN( +inputEl.value)) { // check only numbers
          alert(`Only numbers can be entered`);
        } else { // add item
          addItemToListOfSelectedItems( selectedItemsListEl, inputEl.value,
            `> ${inputEl.value}`, "added");
          inputEl.value = "";
        }
      }
      inputEl.value = "";
      inputEl.focus();
    }
  });
}
/**
 * Fill the select element of an Multiple Choice Widget with option elements created
 * from the selectionRange minus an optional selection set specified in optPar
 *
 * @param {object} selectionRange  An map of objects
 * @param {object} selectEl  A select(ion list) element
 * @param {string} keyProp  The standard identifier property
 * @param {object} optPar [optional]  An record of optional parameter slots
 *                 including optPar.displayProp and optPar.selection
 */
function fillMultipleChoiceWidgetWithOptions(selectEl, selectionRange, keyProp, optPar) {
  // delete old contents
  selectEl.innerHTML = "";
  // create "no selection yet" entry
  selectEl.add( createOption(""," --- "));
  // create option elements
  for (const i of selectionRange) {
    const found = optPar.selection.some(a => a[keyProp] === i[keyProp]);
    // if invoked with a selection argument, only add options for items
    // that are not yet selected
    if (!found) {
      selectEl.add(createOption(i[keyProp], i[optPar.displayProp]));
    }
  }
}
/**
 * Fill a Choice Set element with items
 *
 * @param {object} listEl  A list element
 * @param {object} selection  An entity table for filling the Choice Set
 * @param {string} keyProp  The standard ID property of the entity table
 * @param {string} displayProp  A text property of the entity table
 */
function fillChoiceSet( listEl, selection, keyProp, displayProp) {
  var options = [], obj = null;
  // delete old contents
  listEl.innerHTML = "";
  // create list items from object property values
  options = Object.keys( selection);
  for (const j of options.keys()) {
    obj = selection[options[j]];
    addItemToChoiceSet( listEl, obj[keyProp], obj[displayProp]);
  }
}

/**
 * Add an item to a list element showing selected objects
 *
 * @param {object} listEl  A list element
 * @param {string} stdId  A standard identifier of an object
 * @param {string} humanReadableId  A human-readable ID of the object
 * @param {string} classValue?  A class value to be assigned to the list item
 */
function addItemToListOfSelectedItems( listEl, stdId, humanReadableId, classValue) {
  var el=null;
  const listItemEl = document.createElement("li");
  listItemEl.setAttribute("data-value", stdId);
  el = document.createElement("span");
  el.textContent = humanReadableId;
  listItemEl.appendChild( el);
  el = createPushButton("✕");
  listItemEl.appendChild( el);
  if (classValue) listItemEl.classList.add( classValue);
  listEl.appendChild( listItemEl);
}

/**
 * Add an item to a Choice Set element
 *
 * @param {object} listEl  A list element
 * @param {string} stdId  A standard identifier of an object
 * @param {string} humanReadableId  A human-readable ID of the object
 */
function addItemToChoiceSet(listEl, stdId, humanReadableId, classValue) {
  var listItemEl = null, el = null;
  listItemEl = document.createElement("li");
  listItemEl.setAttribute("data-value", stdId);
  el = document.createElement("span");
  el.textContent = humanReadableId;
  listItemEl.appendChild( el);
  el = createPushButton("✕");
  listItemEl.appendChild( el);
  if (classValue) listItemEl.classList.add( classValue);
  listEl.appendChild( listItemEl);
}
/**
 * Convert Firestore timeStamp object to Date string in format YYYY-MM-DD
 * @param {object} timeStamp A Firestore timeStamp object
 */
function timestampToDate (timeStamp) {
  let dateObj = timeStamp.toDate();
  let  y = dateObj.getFullYear(),
    m = "" + (dateObj.getMonth() + 1),
    d = "" + dateObj.getDate();
  if (m.length < 2) m = "0" + m;
  if (d.length < 2) d = "0" + d;
  return [y, m, d].join("-");
}
/**
 * Convert Date string in format YYYY-MM-DD to a Firestore timeStamp object
 * @param {string} date A date string
 */
function dateToTimestamp (date) {
  return firebase.firestore.Timestamp.fromDate( new Date(date));
}
/**
 * Show or hide progress bar element
 * @param {object} elementId  A node object
 * @param {string} status
 */
function showProgressBar (elementId, status) {
  const section = document.querySelector( `section#${elementId}`),
    progressEl = section.querySelector( 'progress');
  if (status === "show") progressEl.hidden = false;
  if (status === "hide") progressEl.hidden = true;
}
/**
 * Collect all selected option values in a list
 * @param {object} selectEl  A node object
 */
function getValuerFromSelectedOptions (selectEl) {
  let listOfValues = [];
  let selectedOptions = selectEl.selectedOptions;
  for (let opt of selectedOptions) {
    listOfValues.push( parseInt(opt.value));
  }
  return listOfValues;
}

export { fillSelectWithOptions, createListFromMap, createMultiSelectionWidget,
  isNonEmptyString, timestampToDate, dateToTimestamp, nextYear, showProgressBar };
